'''
Contains Focus Tracking Wizard Class: autogenerates start end end foci from reference / anchor positions

'''

from PyQt5 import QtWidgets, QtGui, QtCore

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtProperty

from ..mesoSPIM_State import mesoSPIM_StateSingleton

class FocusTrackingWizard(QtWidgets.QWizard):
    '''
    Wizard to run

    The parent is the Window class of the microscope
    '''
    wizard_done = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        ''' By an instance variable, callbacks to window signals can be handed
        through '''
        self.parent = parent
        self.cfg = parent.cfg
        self.state = mesoSPIM_StateSingleton()

        self.f_1 = 0
        self.f_2 = 0
        self.z_1 = 0
        self.z_2 = 0
        
        self.setWindowTitle('Foucs Tracking Wizard')

        self.addPage(FocusTrackingWizardWelcomePage(self))
        self.addPage(FocusTrackingWizardSetReferencePointsPage(self))
        self.addPage(FocusTrackingWizardCheckResultsPage(self))
        
        self.show()

    def done(self, r):
        ''' Reimplementation of the done function

        if r == 0: canceled
        if r == 1: finished properly
        '''
        if r == 0:
            print("Wizard was canceled")
        if r == 1:
            print('Wizard was closed properly')
            # print('Laser selected: ', self.field('Laser'))
            self.update_focus_positions_in_model()
        else:
            print('Wizard provided return code: ', r)

        super().done(r)

    def calculate_f_pos(self, z_1, z_2, f_1, f_2, z):
        if z_2 == z_1:
            ''' Avoid division by zero '''
            return 0
        else:
            return (f_2-f_1)/(z_2-z_1)*(z-z_1)+f_1

    def convert_string_to_list(self, inputstring):
        outputlist = []
        for substring in inputstring.split(','):
            newlist = substring.split('-')
            if len(newlist) == 1:
                output_values = [int(newlist[0])]
            else:
                output_values = [i for i in range(int(newlist[0]), int(newlist[1])+1, 1)]
            outputlist.extend(output_values)
        return outputlist       

    
    def update_focus_positions_in_model(self):
        row_count = self.parent.model.rowCount()
        z_start_column =  self.parent.model.getColumnByName('Z_start')
        z_end_column = self.parent.model.getColumnByName('Z_end')
        f_start_column = self.parent.model.getColumnByName('F_start')
        f_end_column = self.parent.model.getColumnByName('F_end')
        filter_column = self.parent.model.getColumnByName('Filter')
        laser_column = self.parent.model.getColumnByName('Laser')

        if self.field('RowEnabled'):
            row_list = self.convert_string_to_list(self.field('RowString'))

        for row in range(0, row_count):
            z_start = self.parent.model.getZStartPosition(row)
            z_end = self.parent.model.getZEndPosition(row)

            f_start = self.calculate_f_pos(self.z_1, self.z_2, self.f_1, self.f_2, z_start)
            f_end = self.calculate_f_pos(self.z_1, self.z_2, self.f_1, self.f_2, z_end)

            f_start_index = self.parent.model.createIndex(row, f_start_column)
            f_end_index = self.parent.model.createIndex(row, f_end_column)

            if self.field('LaserEnabled'):
                print('Laser is enabled')
                if self.field('Laser') == 'All laser lines':
                    print('All laser lines')
                    self.parent.model.setData(f_start_index, f_start)
                    self.parent.model.setData(f_end_index, f_end)
                else: 
                    print('Laserfield: ', self.field('Laser'))
                    print('Laser in row: ', self.parent.model.getLaser(row))
                    if self.field('Laser') == self.parent.model.getLaser(row):
                        self.parent.model.setData(f_start_index, f_start)
                        self.parent.model.setData(f_end_index, f_end)

            elif self.field('FilterEnabled'):
                if self.field('Filter') == 'All filters':
                    self.parent.model.setData(f_start_index, f_start)
                    self.parent.model.setData(f_end_index, f_end)
                else:
                    if self.field('Filter') == self.parent.model.getFilter(row):
                        self.parent.model.setData(f_start_index, f_start)
                        self.parent.model.setData(f_end_index, f_end)

            elif self.field('RowEnabled'):
                if row in row_list:
                    self.parent.model.setData(f_start_index, f_start)
                    self.parent.model.setData(f_end_index, f_end)
            

class FocusTrackingWizardWelcomePage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle("Welcome to the focus tracking wizard!")
        self.setSubTitle("This wizard allows you to set the correct focus start and end points by focusing manually at two reference points inside the sample. ATTENTION: In the last step, you can apply the focus range to selected channels and lasers!")
    
class FocusTrackingWizardSetReferencePointsPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle("Reference point definition")
        self.setSubTitle("Between z_start and z_end, focus the microscope at two different z-positions inside the sample.")

        self.button0 = QtWidgets.QPushButton(self)
        self.button0.setText('Set first reference point')
        self.button0.setCheckable(True)
        self.button0.toggled.connect(self.get_first_reference_point)

        self.button1 = QtWidgets.QPushButton(self)
        self.button1.setText('Set second reference point')
        self.button1.setCheckable(True)
        self.button1.toggled.connect(self.get_second_reference_point)

        self.registerField('set_first_refpoint*',
                            self.button0,
                            )
        self.registerField('set_second_refpoint*',
                            self.button1,
                            )

        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.button0, 0, 0)
        self.layout.addWidget(self.button1, 0, 1)
        self.setLayout(self.layout)

    def get_first_reference_point(self):
        self.parent.f_1 = self.parent.state['position']['f_pos']
        self.parent.z_1 = self.parent.state['position']['z_pos']
        
    def get_second_reference_point(self):
        self.parent.f_2 = self.parent.state['position']['f_pos']
        self.parent.z_2 = self.parent.state['position']['z_pos']

    def validatePage(self):
        '''Further validation operations can be introduced here'''
        return super().validatePage()

class FocusTrackingWizardCheckResultsPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle('To which acquisitions should the focus settings be applied?')

        self.laserCheckBox = QtWidgets.QCheckBox('Rows with laser set to: ', self)
        self.laserComboBox = QtWidgets.QComboBox(self)
        self.laserComboBox.addItem('All laser lines')
        self.laserComboBox.addItems(self.parent.cfg.laserdict.keys())

        self.filterCheckBox = QtWidgets.QCheckBox('Rows with filter set to:', self)
        self.filterComboBox = QtWidgets.QComboBox(self)
        self.filterComboBox.addItem('All filters')
        self.filterComboBox.addItems(self.parent.cfg.filterdict.keys())

        self.rowCheckBox = QtWidgets.QCheckBox('Specific rows (e.g. 0-2,5-7,10,13):',self)
        self.rowLineEdit = QtWidgets.QLineEdit(self)
        
        self.registerField('LaserEnabled',self.laserCheckBox)
        self.registerField('FilterEnabled', self.filterCheckBox)
        self.registerField('RowEnabled', self.rowCheckBox)
        self.registerField('Laser',self.laserComboBox, 'currentText', self.laserComboBox.currentTextChanged)
        self.registerField('Filter', self.filterComboBox, 'currentText', self.filterComboBox.currentTextChanged)
        self.registerField('RowString', self.rowLineEdit)

        self.laserCheckBox.clicked.connect(lambda boolean: self.laserComboBox.setEnabled(boolean))
        self.laserCheckBox.clicked.connect(lambda boolean: self.filterCheckBox.setChecked(not boolean))
        self.laserCheckBox.clicked.connect(lambda boolean: self.filterComboBox.setEnabled(not boolean))
        self.laserCheckBox.clicked.connect(lambda boolean: self.rowLineEdit.setEnabled(not boolean))
        self.laserCheckBox.clicked.connect(lambda boolean: self.rowCheckBox.setChecked(not boolean))
        
        self.filterCheckBox.clicked.connect(lambda boolean: self.filterComboBox.setEnabled(boolean))
        self.filterCheckBox.clicked.connect(lambda boolean: self.laserCheckBox.setChecked(not boolean))
        self.filterCheckBox.clicked.connect(lambda boolean: self.laserComboBox.setEnabled(not boolean))
        self.filterCheckBox.clicked.connect(lambda boolean: self.rowLineEdit.setEnabled(not boolean))
        self.filterCheckBox.clicked.connect(lambda boolean: self.rowCheckBox.setChecked(not boolean))
        
        self.rowCheckBox.clicked.connect(lambda boolean: self.rowLineEdit.setEnabled(boolean))
        self.rowCheckBox.clicked.connect(lambda boolean: self.laserComboBox.setEnabled(not boolean))
        self.rowCheckBox.clicked.connect(lambda boolean: self.filterComboBox.setEnabled(not boolean))
        self.rowCheckBox.clicked.connect(lambda boolean: self.laserCheckBox.setChecked(not boolean))
        self.rowCheckBox.clicked.connect(lambda boolean: self.filterCheckBox.setChecked(not boolean))

        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.laserCheckBox, 0, 0)
        self.layout.addWidget(self.laserComboBox, 0, 1)
        self.layout.addWidget(self.filterCheckBox, 1, 0)
        self.layout.addWidget(self.filterComboBox, 1, 1)
        self.layout.addWidget(self.rowCheckBox, 2, 0)
        self.layout.addWidget(self.rowLineEdit, 2, 1)
        self.setLayout(self.layout)

    def initializePage(self):
        self.laserComboBox.setCurrentText(self.parent.state['laser'])
        self.filterComboBox.setCurrentText(self.parent.state['filter'])

        self.laserCheckBox.setChecked(True)
        self.filterComboBox.setEnabled(False)
        self.rowLineEdit.setEnabled(False)






        
    


from PyQt5 import QtWidgets, QtGui, QtCore, QtDesigner

from .acquisitions import Acquisition, AcquisitionList

from ..mesoSPIM_State import mesoSPIM_StateSingleton

import copy
import pickle


class AcquisitionModel(QtCore.QAbstractTableModel):
    '''
    Model class containing a AcquisitionList

    The headers are derived from the keys of the first acquisition
    dictionary.

    TODO: Typecheck in __init__ for AcquisitionList as table
    '''
    def __init__(self, table=None, parent=None):
        super().__init__(parent)
        
        if table is None:
            self._table = AcquisitionList()
        else:
            self._table = table

        ''' Get the headers as the capitalized keys from the first acquisition '''
        self._headers = self._table.get_capitalized_keylist()

        self.state = mesoSPIM_StateSingleton()

        self.dataChanged.connect(self.updatePlanes)

    def rowCount(self, parent = QtCore.QModelIndex()):
        ''' Tells the view how many items this model contains '''
        return len(self._table)

    def columnCount(self, parent = QtCore.QModelIndex()):
        return len(self._table[0])

    def flags(self, index):
        ''' Return Editability for arbitrary elements
        Drag and drop flags are also set here
        '''

        # What does | actually mean?
        '''
        Attention: Here, it is hardcoded that column 5 (# Planes is not editable) If acquistions.py changes, this will 
        require a change here as well
        '''
        if index.column() == self._headers.index('Planes'):
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled
        else: 
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled

    def headerData(self, section, orientation, role):
        '''
        What to display in the table headers

        Orientation: Row part or column part
        Section: Which index

        '''
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self._headers[section]
            if orientation == QtCore.Qt.Vertical:
                return 'Stack ' + str(section)

    def data(self, index, role):
        ''' Data allows to fetch one item'''

        row = index.row()
        column = index.column()

        if role == QtCore.Qt.EditRole:
            ''' What is shown when the editing (double click)
            is selected: A line edit is transiently created
            '''
            return self._table[row](column)

        if role == QtCore.Qt.DisplayRole:
            ''' What is displayed '''
            return self._table[row](column)

        if role == QtCore.Qt.ToolTipRole:
            ''' Tooltip: Text that is display when mouse hovers '''
            return "Table entry: " + str(self._table[row](column))

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        ''' Method used to write data

        Here, a single table entry is set.

        Has to return True when the change was successful
        '''
        if role == QtCore.Qt.EditRole:
            _ = self.columnCount()
            row = index.row()
            column = index.column()

            try:
                '''
                self._table[row] is the Acquisition object
                out of which a column entry needs to be accessed
                - this can only happen via the key
                - if d is the indexed dict, then
                - d[d.keys()[column]] = new_value allow that
                '''
                self._table[row][self._table[row].keys()[column]] = value
                self.dataChanged.emit(index, index)
                #print('Data changed')
                return True
            except:
                #print('Data NOT changed')
                return False

    def setDataFromState(self, row, state_parameter):
        column = self._table.get_keylist().index(state_parameter)

        if state_parameter in ('x_pos','y_pos','z_pos','f_pos'):
            new_value = round(self.state['position'][state_parameter],2)
        elif state_parameter == 'rot':
            new_value = round(self.state['position']['theta_pos'],2)
        else:
            new_value = self.state[state_parameter]

        index = self.createIndex(row, column)

        self.setData(index, new_value)        

    def updatePlanes(self, index, index2):
        ''' Checks if z_planes need to be updated and updates them'''
        column = index.column()

        keylist = self._table.get_keylist()

        z_start_column = keylist.index('z_start')
        z_end_column = keylist.index('z_end')
        z_step_column = keylist.index('z_step')

        if column in (z_start_column ,z_end_column, z_step_column):
            row = index.row()
            planes = self._table[row].get_image_count()
            planes_column = keylist.index('planes')

            index = self.createIndex(row, planes_column)
            self.setData(index, planes)

    def insertRows(self, position, rows, parent=QtCore.QModelIndex()):
        ''' Method to add entries to the model

        Rows: how many rows are inserted at once.

        The views expect a zero-based index

        self.beginInsertRows(index, first, last) to send signals
        '''
        self.beginInsertRows(parent, position, position + rows - 1)

        for i in range(rows):
            # defaultValues = ['Default' for i in range(self.columnCount(parent))]
            # defaultValues = ['Default',1,'One',50,56]
            self._table.insert(position, Acquisition())

        ''' Sending the required signal '''
        self.endInsertRows()

        return True

    def removeRows(self, position, rows, parent = QtCore.QModelIndex()):
        self.beginRemoveRows(parent, position, position + rows - 1)

        for i in range(rows):
            del self._table[position]

        self.endRemoveRows()
        return True

    def copyRow(self, row):
        ''' Copies a row '''
        old_row = copy.deepcopy(self._table[row])
        self.insertRow(row)
        self._table[row] = old_row
        self.send_data_changed()

    # def supportedDropActions(self):
    #     pass
    #     return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction

    def mimeData(self, indices):
        '''
        Here, the model needs to provide a serialization of the entries
        in a QMimeData object
        '''
        #print('MimeData called')
        mimeData = super().mimeData(indices)

        # print(indices[0].row())
        #
        # mimeData = QtCore.QMimeData()
        mimeData.setText(str(self._table[indices[0].row()](0)))
        return mimeData

    def dropMimeData(self, data, action, row, col, parent):
        '''
        Here, the model needs to deserialize the entries
        in a QMimeData object

        Always move the entire row, and don't allow column "shifting"
        '''
        #print('MimeData dropped')
        if action == QtCore.Qt.MoveAction:
            pass
            #print('Row: ', row)
            #print('Col: ', col)
            #print('Data:', data.text())
            #print('HTML:', data.html())
            # print()
            # self.insertRow(row)
            # print('Data methods: ', dir(data))
            # return super().dropMimeData(data, action, row, col, parent)
        if action == QtCore.Qt.CopyAction:
            #print('Copy action detected.')
            pass

        return True

    def moveRows(self, source_parent, source_row, count, destination_parent, destination_row):
        self.rowsAboutToBeMoved.emit(source_parent, source_row, source_row+count-1, destination_parent, destination_row)

        #print('Moving rows: ', source_row, ' #Rows ', count, ' to destination ', destination_row)

        extracted_list = []
        try:
            ''' Remove the sublist from the list '''
            for i in range(count):
                extracted_list.append(self._table.pop(source_row))

            ''' Go through the sublist and add its items '''
            for i in range(count):
                ''' As the _table was shortened during modification, the destination_row has to
                be adapted for insertion as well. '''
                if destination_row > source_row:
                    self._table.insert(destination_row-count+1,extracted_list.pop(len(extracted_list)-1))
                else:
                    self._table.insert(destination_row,extracted_list.pop(len(extracted_list)-1))

            self.send_data_changed()
            return True
        except:
            #print('moveRows failed')
            return False

        self.rowsMoved.emit(source_parent, source_row, source_row+count-1, destination_parent, destination_row)

    def send_data_changed(self):
        ''' Helper method that allows to send a dataChanged Signal for the whole table'''
        top_left_index = self.createIndex(0,0)
        bottom_right_index = self.createIndex(self.rowCount(),self.columnCount())
        self.dataChanged.emit(top_left_index,bottom_right_index)

    def getName(self, row):
        ''' Here, I assume that the filename is the name of the process'''
        return self._table[row]['filename']

    def getFilenameColumn(self):
        return self._headers.index('Filename')

    def getFilename(self, row):
        return self._table[row]['filename']

    def getStartFocusColumn(self):
        return self._headers.index('F_start')

    def getEndFocusColumn(self):
        return self._headers.index('F_end')

    def getColumnByName(self, name):
        return self._headers.index(name)

    def getTime(self, row):
        return int(self._table[row].get_acquisition_time())

    def getFilter(self, row):
        return self._table[row]['filter']

    def getLaser(self, row):
        return self._table[row]['laser']

    def getShutterconfig(self, row):
        return self._table[row]['shutterconfig']

    def getNShutterConfigs(self):
        return self._table.get_n_shutter_configs()

    def getZoom(self, row):
        return self._table[row]['zoom']

    def getImageCount(self, row):
        return self._table[row].get_image_count()

    def getXPosition(self, row):
        return self._table[row]['x_pos']

    def getYPosition(self, row):
        return self._table[row]['y_pos']

    def getZStartPosition(self, row):
        return self._table[row]['z_start']

    def getZEndPosition(self, row):
        return self._table[row]['z_end']

    def getRotationPosition(self, row):
        return self._table[row]['rot']

    def getTotalImageCount(self):
        ''' gets the total number of planes from the model '''
        return self._table.get_image_count()

    def getTileIndex(self, row):
        return self._table.get_tile_index(self._table[row])

    def get_acquisition_list(self, row=None):
        if row is None:
            return self._table
        else:
            return AcquisitionList([self._table[row]])

    def saveModel(self, filename):
        ''' Allows to serialize a model via pickle '''
        with open(filename, "wb" ) as file:
            pickle.dump(self._table, file)

    def setTable(self, table):
        self.modelAboutToBeReset.emit()
        self._table = table
        self.modelReset.emit()

    def loadModel(self, filename):
        self.modelAboutToBeReset.emit()
        with open(filename, "rb" ) as file:
            self._table = pickle.load(file)
        self.modelReset.emit()

    def deleteTable(self):
        self.modelAboutToBeReset.emit()
        self._table = AcquisitionList()
        self.modelReset.emit()



'''
Frontend of the optimizer which allows to use auto-focus or optimization of other microscope parameters
Auto-focus is based on Autopilot paper (Royer at al, Nat Biotechnol. 2016 Dec;34(12):1267-1278. doi: 10.1038/nbt.3708.)
author: Nikita Vladimirov, @nvladimus, 2021
License: GPL-3
'''

import time
import numpy as np
from .utils.optimization import shannon_dct, fit_gaussian_1d, gaussian_1d
import pyqtgraph as pg

import logging
logger = logging.getLogger(__name__)

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.uic import loadUi
from .mesoSPIM_State import mesoSPIM_StateSingleton


class mesoSPIM_Optimizer(QtWidgets.QWidget):
    sig_state_request = QtCore.pyqtSignal(dict)
    sig_move_absolute = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        '''Parent must be an mesoSPIM_MainWindow() object'''
        super().__init__()
        self.parent = parent # the mesoSPIM_MainWindow() object
        self.core = parent.core
        self.cfg = parent.cfg # initial config file
        self.state = mesoSPIM_StateSingleton() # current state
        self.results_window = self.graphics_widget = None
        self.modes_list = ['etl_offset', 'etl_amp', 'focus']
        self.mode = self.modes_list[0]
        self.n_points = 5
        self.search_amplitude = 0.5
        self.state_key = None
        self.image = self.roi = self.roi_dims = self.img_subsampling = None
        self.ini_state = self.ini_metric = self.new_state = self.min_value = self.max_value = None
        self.search_grid = self.metric_array = self.fit_grid = self.gaussian_values = None
        self.delay_s = 0.5  # time delay between snaps to avoid state update hickups, esp for heavy lens-camera assembly during AF

        loadUi('gui/mesoSPIM_Optimizer.ui', self)
        self.setWindowTitle('mesoSPIM-Optimizer')
        self.show()

        # initialize
        #self.set_parameters({'mode': 'etl_offset', 'amplitude': 0.5, 'n_points': 7})

        # signal switchboard
        self.core.camera_worker.sig_camera_frame.connect(self.set_image)
        self.runButton.clicked.connect(self.run_optimization)
        self.closeButton.clicked.connect(self.close_window)

        self.parent.camera_window.sig_update_roi.connect(self.get_roi_dims)
        self.sig_move_absolute.connect(self.core.move_absolute)
        self.comboBoxMode.currentTextChanged.connect(self.set_mode_from_gui)

    @QtCore.pyqtSlot(np.ndarray)
    def set_image(self, image):
        self.image = image
        self.roi = self.image[self.roi_dims[1]:self.roi_dims[1] + self.roi_dims[3],
                   self.roi_dims[0]:self.roi_dims[0] + self.roi_dims[2]]
        # DEBUG mode, create new window for each snap
        #roi_window = pg.ImageWindow()
        #roi_window.setImage(self.roi)

    @QtCore.pyqtSlot(tuple)
    def get_roi_dims(self, roi_dims):
        self.roi_dims = np.array(roi_dims).clip(min=0).astype(int)

    def set_roi(self, orientation='h', roi_perc=0.25):
        img_w, img_h = self.parent.camera_window.get_image_shape()
        if orientation == 'h':
            self.parent.camera_window.set_roi('box', (0, img_h*(1-roi_perc)//2, img_w, int(img_h*roi_perc)))
        elif orientation == 'v':
            self.parent.camera_window.set_roi('box', (img_w*(1-roi_perc)//2, 0, int(img_w*roi_perc), img_h))
        elif orientation == 'c':
            self.parent.camera_window.set_roi('box', (img_h * (1 - roi_perc) // 2, (img_w * (1 - roi_perc)) // 2,
                                                    int(img_h * roi_perc), int(img_w * roi_perc)))
        elif orientation is None:
            self.parent.camera_window.set_roi(None, (0, 0, img_w, img_h))
        else:
            raise ValueError("Orientation must be one of ('h', 'v', None).")

    def set_parameters(self, param_dict=None, update_gui=True):
        if param_dict:
            if 'mode' in param_dict.keys():
                self.mode = param_dict['mode']
            if 'amplitude' in param_dict.keys():
                self.search_amplitude = param_dict['amplitude']
            if 'n_points' in param_dict.keys():
                self.n_points = param_dict['n_points']
        if self.mode == 'focus':
            self.state_key = 'position'
        elif self.mode == 'etl_offset':
            assert self.state['shutterconfig'] in ('Left', 'Right'),  f"Shutter config must be in ('Left', 'Right'), got {self.state['shutterconfig']}"
            self.state_key = 'etl_l_offset' if self.state['shutterconfig'] == 'Left' else 'etl_r_offset'
        elif self.mode == 'etl_amp':
            ini_etl_amp = 0.1 # so that we never start from zero
            self.state_key = 'etl_l_amplitude' if self.state['shutterconfig'] == 'Left' else 'etl_r_amplitude'
            if self.state[self.state_key] == 0:
                self.core.sig_state_request.emit({self.state_key: ini_etl_amp})
                print(f"Initial ETL amp set to {ini_etl_amp}")
        if update_gui:
            self.update_gui()

    def update_gui(self):
        if self.mode == 'focus':
            self.searchAmpDoubleSpinBox.setSuffix(" \u03BCm")
            self.searchAmpDoubleSpinBox.setDecimals(0)
            self.set_roi('c')
        else:
            self.searchAmpDoubleSpinBox.setSuffix(" V")
            self.searchAmpDoubleSpinBox.setDecimals(3)
            self.set_roi('v') if self.mode == 'etl_offset' else self.set_roi('h')

        mode_index = self.modes_list.index(self.mode)
        if mode_index != self.comboBoxMode.currentIndex():
            self.comboBoxMode.setCurrentIndex(mode_index)
        if self.search_amplitude != self.searchAmpDoubleSpinBox.value():
            self.searchAmpDoubleSpinBox.setValue(self.search_amplitude)
        if self.n_points != self.nPointsSpinBox.value():
            self.nPointsSpinBox.setValue(self.n_points)

    @QtCore.pyqtSlot(str)
    def set_mode_from_gui(self, choice):
        if choice == "ETL offset":
            self.set_parameters({'mode': 'etl_offset', 'amplitude': 0.2, 'n_points': 7})
        elif choice == "ETL amplitude":
            self.set_parameters({'mode': 'etl_amp', 'amplitude': 0.1, 'n_points': 7})
        elif choice == "Focus":
            self.set_parameters({'mode': 'focus', 'amplitude': 300, 'n_points': 7})
        else:
            raise ValueError(f"{choice} value is not allowed.")

    def get_params_from_gui(self):
        self.mode = self.modes_list[self.comboBoxMode.currentIndex()]
        self.search_amplitude = self.searchAmpDoubleSpinBox.value()
        self.n_points = self.nPointsSpinBox.value()

    def set_state(self, new_val):
        if self.mode == 'focus':
            self.sig_move_absolute.emit({'f_abs': new_val})
        else:
            self.core.sig_state_request.emit({self.state_key: new_val})

    def set_etl_amp_to_zero(self):
        if self.state_key == 'etl_l_offset':
            self.core.sig_state_request.emit({'etl_l_amplitude': 0})
            print("ETL offset optimization: ETL amplitude (L) set to 0")
        elif self.state_key == 'etl_r_offset':
            self.core.sig_state_request.emit({'etl_r_amplitude': 0})
            print("ETL offset optimization: ETL amplitude (R) set to 0")

    @QtCore.pyqtSlot()
    def run_optimization(self):
        self.parent.sig_state_request.emit({'state': 'idle'}) # stop Live if it is running
        time.sleep(1)
        self.get_params_from_gui()
        self.set_etl_amp_to_zero()
        self.ini_state = self.state[self.state_key]['f_pos'] if self.mode == 'focus' else self.state[self.state_key]
        self.min_value = self.ini_state - self.search_amplitude
        if self.mode in ('etl_offset', 'etl_amp'):
            self.min_value = max(self.min_value, 0) # clip negative values for ETL
        self.max_value = self.ini_state + self.search_amplitude
        assert self.n_points % 2 == 1, f"Number of points must be odd, got {self.n_points} instead."
        self.search_grid = np.linspace(self.min_value, self.max_value, self.n_points)
        self.img_subsampling = self.core.camera_worker.camera_display_acquisition_subsampling
        self.metric_array = np.zeros(len(self.search_grid))
        print(f"Image subsampling: {self.img_subsampling}")
        print(f"Initial value: {self.ini_state:.3f}, searching in ({self.min_value:.3f}, {self.max_value:.3f}), n_points {self.n_points}")
        for i, v in enumerate(self.search_grid):
            self.set_state(v)
            time.sleep(self.delay_s)
            self.core.snap(write_flag=False, laser_blanking=True) # this shares downsampled image via slot self.set_image()
            self.metric_array[i] = shannon_dct(self.roi)

        self.set_state(self.ini_state) # Reset to initial state
        time.sleep(self.delay_s) # give it some time to settle
        self.core.snap(write_flag=False)  # this shares downsampled image via slot self.set_image()
        self.ini_metric = shannon_dct(self.roi)

        #fit with Gaussian
        fit_center, fit_sigma, fit_amp, fit_offset = fit_gaussian_1d(self.metric_array, self.search_grid)
        self.fit_grid = np.linspace(min(self.search_grid), max(self.search_grid), 51)
        self.gaussian_values = gaussian_1d(self.fit_grid, fit_center, fit_sigma, fit_amp, fit_offset)
        self.new_state = fit_center

        # Plot the results
        self.create_results_window()
        self.plot_results()

    def create_results_window(self):
        self.results_window = loadUi('gui/mesoSPIM_Optimizer_Results.ui')
        self.results_window.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.results_window.setWindowTitle('Optimization results')
        # signal switchboard for results window
        self.results_window.acceptButton.clicked.connect(self.accept_new_state)
        self.results_window.discardButton.clicked.connect(self.discard_new_state)

    def plot_results(self):
        layout = pg.QtGui.QGridLayout()
        self.results_window.setLayout(layout)

        self.graphics_widget = pg.GraphicsLayoutWidget(show=True)
        plot0 = self.graphics_widget.addPlot(title='Image metric')
        plot0.addLegend(offset=(0, 0))
        plot0.plot(self.search_grid, self.metric_array, pen=None, symbolBrush=(150,0,150), name='measured')
        plot0.plot(x=[self.ini_state], y=[self.ini_metric],  symbolBrush=(0,0,250), name='old state')
        plot0.plot(x=[self.new_state], y=[max(self.gaussian_values)], symbolBrush=(250, 0, 0), name='new state')
        plot0.plot(self.fit_grid, self.gaussian_values, pen=(200, 0, 0), symbol=None, name='fitted')

        labelStyle = {'color': '#FFF', 'font-size': '14pt'}
        plot0.setLabel('bottom', self.state_key, **labelStyle)
        plot0.setLabel('left', 'Shannon(DCT), AU', **labelStyle)
        plot0.showGrid(x=True)
        plot0.setYRange(min(self.gaussian_values.min(), self.metric_array.min())*0.7,
                        max(self.gaussian_values.max(), self.metric_array.max())*1.2)
        self.results_window.label_results.setText(self.results_string())

        layout.addWidget(self.graphics_widget, 0, 0, 1, 2, QtCore.Qt.AlignHCenter)
        layout.addWidget(self.results_window.label_results, 1, 0, 1, 2, QtCore.Qt.AlignHCenter)
        layout.addWidget(self.results_window.acceptButton, 2, 0, 1, 1, QtCore.Qt.AlignHCenter)
        layout.addWidget(self.results_window.discardButton, 2, 1, 1, 1, QtCore.Qt.AlignHCenter)
        self.results_window.show()

    def results_string(self):
        """Pretty formatting"""
        if self.mode == 'focus':
            return f"Old: {self.ini_state:.0f}\t New: {self.new_state:.0f}\t Diff: {(self.new_state - self.ini_state):.0f}"
        else:
            return f"Old: {self.ini_state:.3f}\t New: {self.new_state:.3f}\t Diff: {(self.new_state - self.ini_state):.3f}"

    @QtCore.pyqtSlot()
    def accept_new_state(self):
        self.set_state(self.new_state)
        print(f"Fitted value: {self.new_state:.3f}")
        time.sleep(self.delay_s)
        self.core.snap(write_flag=False)
        state_str = f"{self.state[self.state_key]}" if self.mode == 'focus' else f"{self.state[self.state_key]:.3f}"
        print(f"New {self.state_key}:{state_str}")
        self.results_window.deleteLater()
        self.results_window = None

    @QtCore.pyqtSlot()
    def discard_new_state(self):
        self.new_state = None
        self.results_window.deleteLater()
        self.results_window = None

    @QtCore.pyqtSlot()
    def close_window(self):
        self.parent.camera_window.set_roi(None)
        if self.results_window:
            self.results_window.deleteLater()
        self.close()



'''
mesoSPIM_control.py
========================================
The core module of the mesoSPIM software
'''

__authors__ = "Fabian Voigt, Nikita Vladimirov"
__license__ = "GPL v3"
__version__ = "1.8.0"

import time
import logging
import argparse
import glob
import os
import sys
import importlib.util
from PyQt5 import QtWidgets

package_directory = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(package_directory)) # this is critical for 'from mesoSPIM.src.mesoSPIM_MainWindow import mesoSPIM_MainWindow' to work in both script and package form.
LOGGING_LEVEL = 'INFO' # 'DEBUG' for fuller info
''' Configuring the logging module before doing anything else'''
timestr = time.strftime("%Y%m%d-%H%M%S")
logging_filename = os.path.join(package_directory, 'log', timestr + '.log')
logging.basicConfig(filename=logging_filename, level=LOGGING_LEVEL,
                    format='%(asctime)-8s:%(levelname)s:%(thread)d:%(module)s:%(funcName)s:%(message)s')
logger = logging.getLogger(__name__)
logger.info('mesoSPIM-control started')

from mesoSPIM.src.mesoSPIM_MainWindow import mesoSPIM_MainWindow

logger.info('Modules loaded')

def load_config_UI(current_path):
    '''
    Bring up a GUI that allows the user to select a microscope configuration to import
    '''
    cfg_app = QtWidgets.QApplication(sys.argv)
    current_path = os.path.abspath('./config')
    global_config_path = ''
    global_config_path , _ = QtWidgets.QFileDialog.getOpenFileName(None,
                                                                   'Open microscope configuration file',current_path)

    if global_config_path != '':
        config = load_config_from_file(global_config_path)
        return config
    else:
        ''' Application shutdown '''
        warning = QtWidgets.QMessageBox.warning(None, 'Shutdown warning',
                                                'No configuration file selected - shutting down!',
                                                QtWidgets.QMessageBox.Ok)
        sys.exit()

def load_config_from_file(path_to_config):
    '''
    Load a microscope configuration from a file using importlib
    '''
    spec = importlib.util.spec_from_file_location('module.name', path_to_config)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)
    logger.info(f'Configuration file loaded: {path_to_config}')
    return config

def stage_referencing_check(cfg):
    '''
    Due to problems with some PI stages loosing reference information
    after restarting the mesoSPIM software, some stage configurations require
    a reference movement to be carried out before starting the rest of the software.

    As reference movements can damage the instrument, this function warns users
    about this problem by message boxes and asks them to reach a safe state.
    '''
    if cfg.stage_parameters['stage_type'] == 'PI_rotzf_and_Galil_xy' or cfg.stage_parameters['stage_type'] == 'PI_rotz_and_Galil_xyf':
        warning = QtWidgets.QMessageBox.warning(None,'Sample z reference movement necessary!',
                'Please move the XYZ stage to position where a reference z movement (to the midpoint of the movement range) is safe!',
                QtWidgets.QMessageBox.Cancel | QtWidgets.QMessageBox.Ok)
        if warning == QtWidgets.QMessageBox.Cancel:
            shutdown_message = QtWidgets.QMessageBox.warning(None,'Shutdown warning',
                    'No reference movement - shutting down!',
                    QtWidgets.QMessageBox.Ok)
            sys.exit()
        else:
            return True
    else:
        return True

def get_parser():
    """
    Parse command-line input arguments
    :return: The argparse parser object
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-C', '--console', action='store_true',  # store_true makes it False by default
                        help='Start a ipython console')
    parser.add_argument('-D', '--demo', action='store_true',
                        help='Start in demo mode')
    return parser
  
def dark_mode_check(cfg, app):
    if (hasattr(cfg, 'dark_mode') and cfg.dark_mode) or (hasattr(cfg, 'ui_options') and cfg.ui_options['dark_mode']):
        import qdarkstyle
        app.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='pyqt5'))

def main(embed_console=False, demo_mode=False):
    """
    Load a configuration file according to the following rules:
    1. If the user asked for demo mode, load the `demo_config.py` file
    2. Else, ff the user did not ask for demo mode:
     - if there is only one non-demo config file, load that.
     - if there are multiple config files, bring up the UI loader.
    """
    print('Starting control software')
    logging.info('mesoSPIM Program started.')
    demo_fname = os.path.join(package_directory, 'config', 'demo_config.py')
    if not os.path.exists(demo_fname):
        raise ValueError(f"Demo file not found: {demo_fname}")

    if demo_mode:
            cfg = load_config_from_file(demo_fname)
            print(f"Loaded config from demo file: {demo_fname}")
    else:
        all_configs = glob.glob(os.path.join(package_directory, 'config', '*.py')) # All possible config files
        all_configs_no_demo = list(filter(lambda f: str.find(f, 'demo_') < 0, all_configs))
        if len(all_configs_no_demo) == 0:
            cfg = load_config_from_file(demo_fname)
            print(f"Loaded config from demo file: {demo_fname}")
        elif len(all_configs_no_demo) == 1:
            config_fname = os.path.join(package_directory, all_configs_no_demo[0])
            cfg = load_config_from_file(config_fname)
            print(f"Loaded config from {config_fname}")
        else:
            cfg = load_config_UI(os.path.join(package_directory, 'config'))

    app = QtWidgets.QApplication(sys.argv)
    dark_mode_check(cfg, app)
    stage_referencing_check(cfg)
    ex = mesoSPIM_MainWindow(package_directory, cfg, "mesoSPIM Main Window, v. " + __version__)
    ex.show()

    # hook up the log display widget
    logging.getLogger().addHandler(ex.log_display_handler)

    if embed_console:
        from traitlets.config import Config
        cfg = Config()
        cfg.InteractiveShellApp.gui = 'qt5'
        import IPython
        IPython.start_ipython(config=cfg, argv=[], user_ns=dict(mSpim=ex, app=app))
    else:
        sys.exit(app.exec_())

def run():
    args = get_parser().parse_args()
    main(embed_console=args.console, demo_mode=args.demo)


if __name__ == '__main__':
    run()



"""
mesoSPIM Module for controlling a discrete zoom changer

Authors: Fabian Voigt, Nikita Vladimirov
"""

import time
from PyQt5 import QtCore
from .devices.servos.dynamixel.dynamixel import Dynamixel


class DemoZoom(QtCore.QObject):
    def __init__(self, zoomdict):
        super().__init__()
        self.zoomdict = zoomdict

    def set_zoom(self, zoom, wait_until_done=False):
        if zoom in self.zoomdict:
            if wait_until_done:
                time.sleep(1)
   

class DynamixelZoom(Dynamixel):
    def __init__(self, zoomdict, COMport, identifier=1, baudrate=115200):
        super().__init__(COMport, identifier, baudrate)
        self.zoomdict = zoomdict

    def set_zoom(self, zoom, wait_until_done=False):
        """Changes zoom after checking that the commanded value exists"""
        if zoom in self.zoomdict:
            self._move(self.zoomdict[zoom], wait_until_done)
            self.zoomvalue = zoom
        else:
            raise ValueError('Zoom designation not in the configuration')



import numpy as np

'''
mesoSPIM configuration file.

Use this file as a starting point to set up all mesoSPIM hardware by replacing the 'Demo' designations
with real hardware one-by-one. Make sure to rename your new configuration file to a different filename
(The extension has to be .py).
'''

'''
Dark mode: Renders the UI dark
'''
ui_options = {'dark_mode' : True, # Dark mode: Renders the UI dark if enabled
              'enable_x_buttons' : True, # Here, specific sets of UI buttons can be disabled
              'enable_y_buttons' : True,
              'enable_z_buttons' : True,
              'enable_f_buttons' : True,
              'enable_rotation_buttons' : True,
              'enable_loading_buttons' : True,
              'window_pos': (100, 100), # position of the main window on the screen, top left corner.
              'usb_webcam': True, # open USB web-camera in a separate window
               }
               
'''
Waveform output for Galvos, ETLs etc.
'''
waveformgeneration = 'DemoWaveFormGeneration' # 'DemoWaveFormGeneration' or 'NI'

'''
Card designations need to be the same as in NI MAX, if necessary, use NI MAX
to rename your cards correctly.

The new mesoSPIM configuration (benchtop-inspired) uses one card (PXI6733) and allows up to 4 laser channels.

Physical channels must be connected in certain order:
- 'galvo_etl_task_line' takes Galvo-L, Galvo-R, ETL-L, ETL-R 
(e.g. value 'PXI6259/ao0:3' means Galvo-L on ao0, Galvo-R on ao1, ETL-L on ao2, ETL-R on ao3)
- 'laser_task_line' takes laser modulation, lasers sorted in increasing wavelength order,
(e.g. value 'PXI6733/ao0:7' means '405 nm' connected to ao4, '488 nm' to ao5, etc.)
'''

acquisition_hardware = {'master_trigger_out_line' : 'PXI1Slot4/port0/line0',
                        'camera_trigger_source' : '/PXI1Slot4/PFI0',
                        'camera_trigger_out_line' : '/PXI1Slot4/ctr0',
                        'galvo_etl_task_line' : 'PXI1Slot4/ao0:3',
                        'galvo_etl_task_trigger_source' : '/PXI1Slot4/PFI0',
                        'laser_task_line' :  'PXI1Slot4/ao4:7',
                        'laser_task_trigger_source' : '/PXI1Slot4/PFI0'}

sidepanel = 'Demo' #'Demo' or 'FarmSimulator', deprecated

'''
Digital laser enable lines
'''

laser = 'Demo' # 'Demo' or 'NI'

''' The `laserdict` specifies laser labels of the GUI and their digital modulation channels. 
Keys are the laser designation that will be shown in the user interface
Values are DO ports used for laser ENABLE digital signal.
Critical: entries must be sorted in the increasing wavelength order: 405, 488, etc.
'''
laserdict = {'488 nm': 'PXI1Slot4/port0/line2',
             '520 nm': 'PXI1Slot4/port0/line3',
             '568 nm': 'PXI1Slot4/port0/line4',
             '638 nm': 'PXI1Slot4/port0/line5',
             }


''' Laser blanking indicates whether the laser enable lines should be set to LOW between
individual images or stacks. This is helpful to avoid laser bleedthrough between images caused by insufficient
modulation depth of the analog input (even at 0V, some laser light is still emitted).
'''
laser_blanking = 'images' # if 'images', laser is off before and after every image; if 'stacks', before and after each stack.

'''
Shutter configuration
If shutterswitch = True:
    'shutter_left' is the general shutter
    'shutter_right' is the left/right switch (Right==True)
    
If shutterswitch = False or missing:
    'shutter_left' and 'shutter_right' are two independent shutters.
'''
shutter = 'Demo' # 'Demo' or 'NI'
shutterswitch = True # assumes that the shutter_left line is the general shutter
shutteroptions = ('Left', 'Right') # Shutter options of the GUI
shutterdict = {'shutter_left' : '/PXI1Slot4/port0/line6', # left (general) shutter
              'shutter_right' : '/PXI1Slot4/port0/line1', # flip mirror or right shutter, depending on physical configuration
              }

'''
Camera configuration

For a DemoCamera, only the following options are necessary
(x_pixels and y_pixels can be chosen arbitrarily):

camera_parameters = {'x_pixels' : 1024,
                     'y_pixels' : 1024,
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4]}

For a Hamamatsu Orca Flash 4.0 V2 or V3, the following parameters are necessary:

camera_parameters = {'x_pixels' : 2048,
                     'y_pixels' : 2048,
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4],
                     'camera_id' : 0,
                     'sensor_mode' : 12,    # 12 for progressive
                     'defect_correct_mode': 2,
                     'binning' : '1x1',
                     'readout_speed' : 1,
                     'trigger_active' : 1,
                     'trigger_mode' : 1, # it is unclear if this is the external lightsheeet mode - how to check this?
                     'trigger_polarity' : 2, # positive pulse
                     'trigger_source' : 2, # external
                    }

For a Photometrics Iris 15, the following parameters are necessary:

camera_parameters = {'x_pixels' : 5056, 
                     'y_pixels' : 2960, 
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4],
                     'speed_table_index': 0,
                     'exp_mode' : 'Edge Trigger', # Lots of options in PyVCAM --> see constants.py
                     'readout_port': 0,
                     'gain_index': 1,
                     'exp_out_mode': 4, # 4: line out 
                     'binning' : '1x1',
                     'scan_mode' : 1, # Scan mode options: {'Auto': 0, 'Line Delay': 1, 'Scan Width': 2}
                     'scan_direction' : 0, # Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}
                     'scan_line_delay' : 6, # 10.26 us x factor, a factor = 6 equals 71.82 us
                    }

For a Photometrics Prime BSI Express, the following parameters are necessary:

camera_parameters = {'x_pixels' : 2048, #5056
                     'y_pixels' : 2048, # 2960
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4],
                     'speed_table_index': 1, # 1 for 100 MHz
                     'exp_mode' : 'Edge Trigger', # Lots of options in PyVCAM --> see constants.py
                     'readout_port': 0,
                     'gain_index': 1, # Enable HDR mode
                     'exp_out_mode': 4, # 4: line out 
                     'binning' : '1x1',
                     'scan_mode' : 1, # Scan mode options: {'Auto': 0, 'Line Delay': 1, 'Scan Width': 2}
                     'scan_direction' : 0, # Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}
                     'scan_line_delay' : 7, # 11.2 us x factor, a factor = 3 equals 33.6 us
                    }

'''
camera = 'DemoCamera' # 'DemoCamera' or 'HamamatsuOrcaFlash' or 'Photometrics'

camera_parameters = {'x_pixels' : 1024,
                     'y_pixels' : 1024,
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4],
                     'camera_id' : 0,
                     'sensor_mode' : 12,    # 12 for progressive
                     'defect_correct_mode': 2,
                     'binning' : '1x1',
                     'readout_speed' : 1,
                     'trigger_active' : 1,
                     'trigger_mode' : 1, # it is unclear if this is the external lightsheeet mode - how to check this?
                     'trigger_polarity' : 2, # positive pulse
                     'trigger_source' : 2, # external
                    }

binning_dict = {'1x1': (1,1), '2x2':(2,2), '4x4':(4,4)}

'''
Stage configuration
'''

'''
The stage_parameter dictionary defines the general stage configuration, initial positions,
and safety limits. The rotation position defines a XYZ position (in absolute coordinates)
where sample rotation is safe. Additional hardware dictionaries (e.g. pi_parameters)
define the stage configuration details.
All positions are absolute.

'stage_type' options: 'DemoStage', 'PI_1controllerNstages' (former 'PI'), 'PI_NcontrollersNstages', 'TigerASI'
'''

stage_parameters = {'stage_type' : 'DemoStage', # 'DemoStage'. 'PI_1controllerNstages', 'PI_NcontrollersNstages', 'TigerASI'
                    'y_load_position': -86000,
                    'y_unload_position': -120000,
                    'x_max' : 5000,
                    'x_min' : -5000,
                    'y_max' : 5000,
                    'y_min' : -5000,
                    'z_max' : 5000,
                    'z_min' : -5000,
                    'f_max' : 5000,
                    'f_min' : -5000,
                    'theta_max' : 999,
                    'theta_min' : -999,
                    'x_rot_position': 0,
                    'y_rot_position': -121000,
                    'z_rot_position': 66000,
                    }

''''
If 'stage_type' = 'PI_1controllerNstages' (vanilla mesoSPIM V5 with single 6-axis controller):
pi_parameters = {'controllername' : 'C-884',
                 'stages' : ('L-509.20DG10','L-509.40DG10','L-509.20DG10','M-060.DG','M-406.4PD','NOSTAGE'),
                 'refmode' : ('FRF',),
                 'serialnum' : ('118075764'),
                 }

If 'stage_type' = 'PI_NcontrollersNstages' (mesoSPIM V5 with multiple single-axis controllers):
pi_parameters = {'axes_names': ('x', 'y', 'z', 'theta', 'f'),
                'stages': ('L-509.20SD00', 'L-509.40SD00', 'L-509.20SD00', None, 'MESOSPIM_FOCUS'),
                'controllername': ('C-663', 'C-663', 'C-663', None, 'C-663'),
                'serialnum': ('**********', '**********', '**********', None, '**********'),
                'refmode': ('FRF', 'FRF', 'FRF', None, 'RON')
                }
                
If 'stage_type' = 'TigerASI' (benchtop mesoSPIM with an ASI Tiger controller)
The stage assignment dictionary assigns a mesoSPIM stage (xyzf and theta - dict key) to an ASI stage (XYZ etc) 
which are the values of the dict.
'''

asi_parameters = {'COMport' : 'COM32',
                  'baudrate' : 115200,
                  'stage_assignment': {'x':'X', 'y':'V', 'z':'Z', 'theta':'T', 'f':'Y'},
                  'stage_trigger_source': '/PXI1Slot4/PFI0',
                  'stage_trigger_out_line': '/PXI1Slot4/ctr1',
                  'stage_trigger_delay_%' : 92.5, # Set to 92.5 for stage triggering exactly after the ETL sweep
                  'stage_trigger_pulse_%' : 1,
                  'ttl_motion_enabled': False,
                  }

'''
Filterwheel configuration
For a DemoFilterWheel, no COMport needs to be specified.
For a Ludl Filterwheel, a valid COMport is necessary. Ludl marking 10 = position 0.
For a Dynamixel FilterWheel, valid baudrate and servoi_id are necessary. 
'''
filterwheel_parameters = {'filterwheel_type' : 'Demo', # 'Demo', 'Ludl', 'Dynamixel'
                          'COMport' : 'COM3',
                          'baudrate' : 115200, # relevant only for 'Dynamixel'
                          'servo_id' :  1, # relevant only for 'Dynamixel'
                          }
'''
filterdict contains filter labels and their positions. The valid positions are:
For Ludl: 0, 1, 2, 3, .., 9, i.e. position ids (int)
For Dynamixel: servo encoder counts, e.g. 0 for 0 deg, 1024 for 45 deg (360 deg = 4096 counts, or 11.377 counts/deg). 
Dynamixel encoder range in multi-turn mode: -28672 .. +28672 counts.
'''
filterdict = {'Empty-Alignment' : 0, # Every config should contain this
              '405-488-647-Tripleblock' : 1,
              '405-488-561-640-Quadrupleblock' : 2,
              '464 482-35' : 3,
              '508 520-35' : 4,
              '515LP' : 5,
              '529 542-27' : 6,
              '561LP' : 7,
              '594LP' : 8,
              '417 447-60' : 9}

'''
Zoom configuration
For the DemoZoom, servo_id, COMport and baudrate do not matter. For a Dynamixel zoom,
these values have to be there
'''
zoom_parameters = {'zoom_type' : 'DemoZoom', # 'DemoZoom' or 'Dynamixel'
                   'servo_id' :  4,
                   'COMport' : 'COM38',
                   'baudrate' : 1000000}

'''
The keys in the zoomdict define what zoom positions are displayed in the selection box
(combobox) in the user interface.
'''

zoomdict = {'0.63x' : 3423,
            '0.8x' : 3071,
            '1x' : 2707,
            '1.25x' : 2389,
            '1.6x' : 2047,
            '2x' : 1706,
            '2.5x' : 1354,
            '3.2x' : 967,
            '4x' : 637,
            '5x' : 318,
            '6.3x' : 0}
'''
Pixelsize in micron
'''
pixelsize = {'0.63x' : 10.52,
            '0.8x' : 8.23,
            '1x' : 6.55,
            '1.25x' : 5.26,
            '1.6x' : 4.08,
            '2x' : 3.26,
            '2.5x' : 2.6,
            '3.2x' : 2.03,
            '4x' : 1.60,
            '5x' : 1.27,
            '6.3x' : 1.03}

'''
 HDF5 parameters, if this format is used for data saving (optional).
Downsampling and compression slows down writing by 5x - 10x, use with caution.
Imaris can open these files if no subsampling and no compression is used.
'''
hdf5 = {'subsamp': ((1, 1, 1),), #((1, 1, 1),) no subsamp, ((1, 1, 1), (1, 4, 4)) for 2-level (z,y,x) subsamp.
        'compression': None, # None, 'gzip', 'lzf'
        'flip_xyz': (True, True, False) # match BigStitcher coordinates to mesoSPIM axes.
        }

'''                                                                  
Initial acquisition parameters
Used as initial values after startup
When setting up a new mesoSPIM, make sure that:
* 'max_laser_voltage' is correct (5 V for Toptica MLEs, 10 V for Omicron SOLE)
* 'galvo_l_amplitude' and 'galvo_r_amplitude' (in V) are correct (not above the max input allowed by your galvos)
* all the filepaths exist
* the initial filter exists in the filter dictionary above
'''

startup = {
'state' : 'init', # 'init', 'idle' , 'live', 'snap', 'running_script'
'samplerate' : 100000,
'sweeptime' : 0.2,
'position' : {'x_pos':0,'y_pos':1,'z_pos':2,'f_pos':3,'theta_pos':180},
'ETL_cfg_file' : 'config/etl_parameters/ETL-parameters.csv',
#'filepath' : '/tmp/file.raw', #deprecated
'folder' : 'D:/tmp/',
'snap_folder' : 'D:/tmp/',
'file_prefix' : '',
'file_suffix' : '000001',
'zoom' : '1x',
'pixelsize' : 6.55,
'laser' : '488 nm',
'max_laser_voltage':10,
'intensity' : 10,
'shutterstate':False, # Is the shutter open or not?
'shutterconfig':'Right', # Can be "Left", "Right","Both","Interleaved"
'laser_interleaving':False,
'filter' : '405-488-561-640-Quadrupleblock',
'etl_l_delay_%' : 7.5,
'etl_l_ramp_rising_%' : 85,
'etl_l_ramp_falling_%' : 2.5,
'etl_l_amplitude' : 0.7,
'etl_l_offset' : 2.3,
'etl_r_delay_%' : 2.5,
'etl_r_ramp_rising_%' : 5,
'etl_r_ramp_falling_%' : 85,
'etl_r_amplitude' : 0.65,
'etl_r_offset' : 2.36,
'galvo_l_frequency' : 99.9,
'galvo_l_amplitude' : 2.5,
'galvo_l_offset' : 0,
'galvo_l_duty_cycle' : 50,
'galvo_l_phase' : np.pi/2,
'galvo_r_frequency' : 99.9,
'galvo_r_amplitude' : 2.5,
'galvo_r_offset' : 0,
'galvo_r_duty_cycle' : 50,
'galvo_r_phase' : np.pi/2,
'laser_l_delay_%' : 10,
'laser_l_pulse_%' : 87,
'laser_l_max_amplitude_%' : 100,
'laser_r_delay_%' : 10,
'laser_r_pulse_%' : 87,
'laser_r_max_amplitude_%' : 100,
'camera_delay_%' : 10,
'camera_pulse_%' : 1,
'camera_exposure_time':0.02,
'camera_line_interval':0.000075,
'camera_display_live_subsampling': 2,
#'camera_display_snap_subsampling': 1, #deprecated
'camera_display_acquisition_subsampling': 2,
'camera_binning':'1x1',
'camera_sensor_mode':'ASLM',
'average_frame_rate': 4.969,
}


'''
Contains a variety of mesoSPIM utility functions
'''

import numpy as np

def convert_seconds_to_string(delta_t):
    '''
    Converts an input value in seconds into a string in the format hh:mm:ss

    Interestingly, a variant using np.divmod is around 4-5x slower in initial tests.
    '''
    if delta_t <= 0:
        return '--:--:--'
    else:
        hours, remainder = divmod(delta_t, 3600)
        minutes, seconds = divmod(remainder, 60)
        return f"{int(hours):02}:{int(minutes):02}:{int(seconds):02}"



class Demo_LaserEnabler:
    def __init__(self, laserdict):
        self.laserenablestate = 'None'
        self.laserdict = laserdict

    def _check_if_laser_in_laserdict(self, laser):
        if laser in self.laserdict:
            return True
        else:
            raise ValueError('Laser not in the configuration')
    
    def enable(self, laser):
        if self._check_if_laser_in_laserdict(laser) is True:
            self.laserenablestate = laser
        else:
            pass

    def enable_all(self):
        self.laserenablestate = 'all on'

    def disable(self, laser):
        self.laserenablestate = 'off'

    def disable_all(self):
        self.laserenablestate = 'off'


    def state(self):
        return self.laserenablestate


"""
Python wrapper for Galil gclib.
Contact softwaresupport@galil.com with questions, comments, and suggestions.
"""
###############################################################################
# ctypes import, for pulling in all dll/so/dylib calls. 
# Part of implementation, don't use directly.
###############################################################################
import platform #for distinguishing 'Windows', 'Linux', 'Darwin'
from ctypes import *

if platform.system() == 'Windows':
    if '64 bit' in platform.python_compiler():
        _gclib_path = r'C:\Program Files (x86)\Galil\gclib\dll\x64\gclib.dll'
        _gclibo_path = r'C:\Program Files (x86)\Galil\gclib\dll\x64\gclibo.dll'
        _gclib = WinDLL(_gclib_path)
        _gclibo = WinDLL(_gclibo_path)
    else:
        _gclib_path = r'C:\Program Files (x86)\Galil\gclib\dll\x86\gclib.dll'
        _gclibo_path = r'C:\Program Files (x86)\Galil\gclib\dll\x86\gclibo.dll'
        _gclib = WinDLL(_gclib_path)
        _gclibo = WinDLL(_gclibo_path)
        #Reassign symbol name, Python doesn't like @ in function names
        #gclib calls
        setattr(_gclib, 'GArrayDownload', getattr(_gclib, '_GArrayDownload@20'))
        setattr(_gclib, 'GArrayUpload', getattr(_gclib, '_GArrayUpload@28'))
        setattr(_gclib, 'GClose', getattr(_gclib, '_GClose@4'))
        setattr(_gclib, 'GCommand', getattr(_gclib, '_GCommand@20'))
        setattr(_gclib, 'GFirmwareDownload', getattr(_gclib, '_GFirmwareDownload@8'))
        setattr(_gclib, 'GInterrupt', getattr(_gclib, '_GInterrupt@8'))
        setattr(_gclib, 'GMessage', getattr(_gclib, '_GMessage@12'))
        setattr(_gclib, 'GOpen', getattr(_gclib, '_GOpen@8'))
        setattr(_gclib, 'GProgramDownload', getattr(_gclib, '_GProgramDownload@12'))
        setattr(_gclib, 'GProgramUpload', getattr(_gclib, '_GProgramUpload@12'))
        #gclibo calls (open source component/convenience functions)
        setattr(_gclibo, 'GAddresses', getattr(_gclibo, '_GAddresses@8'))
        setattr(_gclibo, 'GArrayDownloadFile', getattr(_gclibo, '_GArrayDownloadFile@8'))
        setattr(_gclibo, 'GArrayUploadFile', getattr(_gclibo, '_GArrayUploadFile@12'))
        setattr(_gclibo, 'GAssign', getattr(_gclibo, '_GAssign@8'))
        setattr(_gclibo, 'GError', getattr(_gclibo, '_GError@12'))
        setattr(_gclibo, 'GInfo', getattr(_gclibo, '_GInfo@12'))
        setattr(_gclibo, 'GIpRequests', getattr(_gclibo, '_GIpRequests@8'))
        setattr(_gclibo, 'GMotionComplete', getattr(_gclibo, '_GMotionComplete@8'))
        setattr(_gclibo, 'GProgramDownloadFile', getattr(_gclibo, '_GProgramDownloadFile@12'))
        setattr(_gclibo, 'GSleep', getattr(_gclibo, '_GSleep@4'))
        setattr(_gclibo, 'GProgramUploadFile', getattr(_gclibo, '_GProgramUploadFile@8'))
        setattr(_gclibo, 'GTimeout', getattr(_gclibo, '_GTimeout@8'))
        setattr(_gclibo, 'GVersion', getattr(_gclibo, '_GVersion@8'))
        setattr(_gclibo, 'GSetupDownloadFile', getattr(_gclibo, '_GSetupDownloadFile@20'))

elif platform.system() == 'Linux':
    cdll.LoadLibrary("libgclib.so.0")
    _gclib = CDLL("libgclib.so.0")
    cdll.LoadLibrary("libgclibo.so.0")
    _gclibo = CDLL("libgclibo.so.0")

elif platform.system() == 'Darwin': #OSX
    _gclib_path = '/Applications/gclib/dylib/gclib.0.dylib'
    _gclibo_path = '/Applications/gclib/dylib/gclibo.0.dylib'
    cdll.LoadLibrary(_gclib_path)
    _gclib = CDLL(_gclib_path)
    cdll.LoadLibrary(_gclibo_path)
    _gclibo = CDLL(_gclibo_path)
    
    

# Python "typedefs"
_GReturn = c_int #type for a return code
_GCon = c_void_p #type for a Galil connection handle
_GCon_ptr = POINTER(_GCon) #used for argtypes declaration
_GSize = c_ulong #type for a Galil size variable
_GSize_ptr = POINTER(_GSize) #used for argtypes declaration
_GCStringIn = c_char_p #char*. In C it's const.
_GCStringOut = c_char_p #char*
_GOption = c_int #type for option variables, e.g.    GArrayDownload 
_GStatus = c_ubyte #type for interrupt status bytes
_GStatus_ptr = POINTER(_GStatus) #used for argtypes declaration

#Define arguments and result type (if not C int type)
#gclib calls
_gclib.GArrayDownload.argtypes = [_GCon, _GCStringIn, _GOption, _GOption, _GCStringIn]
_gclib.GArrayUpload.argtypes = [_GCon, _GCStringIn, _GOption, _GOption, _GOption, _GCStringOut, _GSize]
_gclib.GClose.argtypes = [_GCon]
_gclib.GCommand.argtypes = [_GCon, _GCStringIn, _GCStringOut, _GSize, _GSize_ptr]
_gclib.GFirmwareDownload.argtypes = [_GCon, _GCStringIn]
_gclib.GInterrupt.argtypes = [_GCon, _GStatus_ptr]
_gclib.GMessage.argtypes = [_GCon, _GCStringOut, _GSize]
_gclib.GOpen.argtypes = [_GCStringIn, _GCon_ptr]
_gclib.GProgramDownload.argtypes = [_GCon, _GCStringIn, _GCStringIn]
_gclib.GProgramUpload.argtypes = [_GCon, _GCStringOut, _GSize]
#gclibo calls (open source component/convenience functions)
_gclibo.GAddresses.argtypes = [_GCStringOut, _GSize]
_gclibo.GArrayDownloadFile.argtypes = [_GCon, _GCStringIn]
_gclibo.GArrayUploadFile.argtypes = [_GCon, _GCStringIn, _GCStringIn]
_gclibo.GAssign.argtypes = [_GCStringIn, _GCStringIn]
_gclibo.GError.argtypes = [_GReturn, _GCStringOut, _GSize]
_gclibo.GError.restype    = None
_gclibo.GError.argtypes = [_GCon, _GCStringOut, _GSize]
_gclibo.GIpRequests.argtypes = [_GCStringOut, _GSize]
_gclibo.GMotionComplete.argtypes = [_GCon, _GCStringIn]
_gclibo.GProgramDownloadFile.argtypes = [_GCon, _GCStringIn, _GCStringIn]
_gclibo.GSleep.argtypes = [c_uint]
_gclibo.GSleep.restype    = None
_gclibo.GProgramUploadFile.argtypes = [_GCon, _GCStringIn]
_gclibo.GTimeout.argtypes = [_GCon, c_int]
_gclibo.GVersion.argtypes = [_GCStringOut, _GSize]
_gclibo.GSetupDownloadFile.argtypes = [_GCon, _GCStringIn, _GOption, _GCStringOut, _GSize]

#Set up some constants
_enc = "ASCII" #byte encoding for going between python strings and c strings.
_buf_size = 500000 #size of response buffer. Big enough to fit entire 4000 program via UL/LS, or 24000 elements of array data.
_error_buf = create_string_buffer(128)    #buffer for retrieving error code descriptions.
    
def _rc(return_code):
    """Checks return codes from gclib and raises a python error if result is exceptional."""
    if return_code != 0:
        _gclibo.GError(return_code, _error_buf, 128) #Get the library's error description
        raise GclibError(str(_error_buf.value.decode(_enc)))
    return 

class GclibError(Exception):
    """Error class for non-zero gclib return codes."""
    pass 
 
class py:
    """Represents a single Python connection to a Galil Controller or PLC."""
    
    def __init__(self):
        """Constructor for the Connection class. Initializes gclib's handle and read buffer."""
        self._gcon = _GCon(0) #handle to connection
        self._buf = create_string_buffer(_buf_size)
        self._timeout = 5000
        return        
    
    def __del__(self):
        """Destructor for the Connection class. Ensures close gets called to release Galil resource (Sockets, Kernel Driver, Com Port, etc)."""
        self.GClose()
        return
    
    def _cc(self):
        """Checks if connection is established, throws error if not."""
        if self._gcon.value == None:
            _rc(-1201) #G_CONNECTION_NOT_ESTABLISHED
    
    def GOpen(self, address):
        """
        Opens a connection a galil controller.
        See the gclib docs for address string formatting.
        See Link GOpen() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_aef4aec8a85630eed029b7a46aea7db54.html#aef4aec8a85630eed029b7a46aea7db54>
        """
        c_address = _GCStringIn(address.encode(_enc))
        _rc(_gclib.GOpen(c_address, byref(self._gcon)))
        return
        
     
    def GClose(self):
        """
        Closes a connection to a Galil Controller. 
        See Link GClose() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_a24a437bcde9637b0db4b94176563a052.html#a24a437bcde9637b0db4b94176563a052>
        """
        if self._gcon.value != None:
            _rc(_gclib.GClose(self._gcon))
            self._gcon = _GCon(0)
        return
        
        
    def GCommand(self, command):
        """
        Performs a command-and-response transaction on the connection. 
        Trims the response.
        See Link GCommand() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_a5ac031e76efc965affdd73a1bec084a8.html#a5ac031e76efc965affdd73a1bec084a8>
        """
        self._cc()
        c_command = _GCStringIn(command.encode(_enc))
        _rc(_gclib.GCommand(self._gcon, c_command, self._buf, _buf_size, None))
        response = str(self._buf.value.decode(_enc))
        return response[:-3].strip() # trim trailing /r/n: and leading space

        
    def GSleep(self, val):
        """
        Provides a blocking sleep call which can be useful for timing-based chores.
        See Link GSleep() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_adf81cb901627dd83408ca1f686e05cac.html#adf81cb901627dd83408ca1f686e05cac>
        """
        _gclibo.GSleep(val)
        return 
        
        
    def GVersion(self):
        """
        Provides the gclib version number. Please include the output of this function on all support cases.
        See Link GVersion() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a1784b39416b77af20efc98a05f8ce475.html#a1784b39416b77af20efc98a05f8ce475>
        """
        _rc(_gclibo.GVersion(self._buf, _buf_size))
        return "py." + str(self._buf.value.decode(_enc))
        
        
    def GInfo(self):
        """
        Provides a useful connection string. Please include the output of this function on all support cases.
        See Link GInfo() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a08abfcff8a1a85a01987859473167518.html#a08abfcff8a1a85a01987859473167518>
        """
        _rc(_gclibo.GInfo(self._gcon, self._buf, _buf_size))
        return str(self._buf.value.decode(_enc))
        
        
    def GIpRequests(self):
        """
        Provides a dictionary of all Galil controllers requesting IP addresses via BOOT-P or DHCP. 
        
        Returns a dictionary mapping 'model-serial' --> 'mac address'
        e.g. {'DMC4000-783': '00:50:4c:20:03:0f', 'DMC4103-9998': '00:50:4c:38:27:0e'}
        
        Linux/OS X users must be root to use GIpRequests() and have UDP access to bind and listen on port 67.
        See Link GIpRequests() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a0afb4c82642a4ef86f997c39a5518952.html#a0afb4c82642a4ef86f997c39a5518952>
        """
        _rc(_gclibo.GIpRequests(self._buf, _buf_size)) #get the c string from gclib
        ip_req_dict = {}
        for line in str(self._buf.value.decode(_enc)).splitlines():
            line = line.replace(' ', '') #trim spaces throughout
            if (line == ""): continue
            fields = line.split(',')
            #fields go [model, serial number, mac]
            ip_req_dict[fields[0] + '-' + fields[1]] = fields[2] # e.g. DMC4000-783 maps to its MAC addr.
        return ip_req_dict
    
    
    def GAssign(self, ip, mac):
        """
        Assigns IP address over the Ethernet to a controller at a given MAC address.
        Linux/OS X users must be root to use GAssign() and have UDP access to send on port 68.
        See Link GAssign() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a80e87e876408dea6e1c5e29da2de143f.html#a80e87e876408dea6e1c5e29da2de143f>
        """
        c_ip = _GCStringIn(ip.encode(_enc))
        c_mac = _GCStringIn(mac.encode(_enc))
        _rc(_gclibo.GAssign(c_ip, c_mac))
        return
        
        
    def GAddresses(self):
        """
        Provides a dictionary of all available connection addresses. 
        
        Returns a dictionary mapping 'address' -> 'revision reports', where possible
        e.g. {}
        
        See Link GAddresses() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8c_ae92a1b09a2f340fc7720ea2074f4526c.html#ae92a1b09a2f340fc7720ea2074f4526c>
        """
        _rc(_gclibo.GAddresses(self._buf, _buf_size))
        addr_dict = {}
        for line in str(self._buf.value.decode(_enc)).splitlines():
            print("LINE ")
            fields = line.split(',')
            if len(fields) >= 2:
                addr_dict[fields[0]] = fields[1]
            else:
                addr_dict[fields[0]] = ''
                
        return addr_dict
 
        
    def GProgramDownload(self, program, preprocessor=""):
        """
        Downloads a program to the controller's program buffer.
        See the gclib docs for preprocessor options.
        See Link GProgramDownload() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_acafe19b2dd0537ff458e3c8afe3acfeb.html#acafe19b2dd0537ff458e3c8afe3acfeb>
        """
        self._cc()
        c_prog = _GCStringIn(program.encode(_enc))
        c_pre = _GCStringIn(preprocessor.encode(_enc))
        _rc(_gclib.GProgramDownload(self._gcon, c_prog, c_pre))
        return
     
    
    def GProgramUpload(self):    
        """
        Uploads a program from the controller's program buffer. 
        See Link GProgramUpload() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_a80a653ce387a2bd16bde2793c6de77e9.html#a80a653ce387a2bd16bde2793c6de77e9>
        """
        self._cc()
        _rc(_gclib.GProgramUpload(self._gcon, self._buf, _buf_size))
        return str(self._buf.value.decode(_enc))
        
        
    def GProgramDownloadFile(self, file_path, preprocessor=""):
        """
        Program download from file. 
        See the gclib docs for preprocessor options.
        See Link GProgramDownloadFile() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a8e44e2e321df9e7b8c538bf2d640633f.html#a8e44e2e321df9e7b8c538bf2d640633f>
        """
        self._cc()
        c_path = _GCStringIn(file_path.encode(_enc))
        c_pre = _GCStringIn(preprocessor.encode(_enc))
        _rc(_gclibo.GProgramDownloadFile(self._gcon, c_path, c_pre))
        return        
        
    def GProgramUploadFile(self, file_path):
        """
        Program upload to file. 
        See Link GProgramUploadFile() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a38c5565afc11762fa19d37fbaa3c9aa3.html#a38c5565afc11762fa19d37fbaa3c9aa3>
        """
        self._cc()
        c_path = _GCStringIn(file_path.encode(_enc))
        _rc(_gclibo.GProgramUploadFile(self._gcon, c_path))
        return
        
    def GArrayDownload(self, name, first, last, array_data):
        """
        Downloads array data to a pre-dimensioned array in the controller's array table. 
        array_data should be a list of values (e.g. int or float)
        See Link GArrayDownload() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_a6ea5ae6d167675e4c27ccfaf2f240f8a.html#a6ea5ae6d167675e4c27ccfaf2f240f8a>
        """
        self._cc()
        c_name = _GCStringIn(name.encode(_enc))
        array_string = ""
        for val in array_data:
            array_string += str(val) + ","
        c_data = _GCStringIn(array_string[:-1].encode(_enc)) #trim trailing command
        _rc(_gclib.GArrayDownload(self._gcon, c_name, first, last, c_data))
        return
        
        
    def GArrayUploadFile(self, file_path, names = []):
        """
        Uploads the entire controller array table or a subset and saves the data as a csv file specified by file_path.
        names is optional and should be a list of array names on the controller.
        See Link GArrayUploadFile() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a2271aa7eff5d1aaf7fc1eb3a472e3f4c.html#a2271aa7eff5d1aaf7fc1eb3a472e3f4c>
        """
        self._cc()
        c_path = _GCStringIn(file_path.encode(_enc))
        names_string = ''
        c_names = _GCStringIn(''.encode(_enc)) #in case empty list provided
        for name in names:
            names_string += name + ' '
        
        c_names = _GCStringIn(names_string[:-1].encode(_enc)) #trim trailing space
        _rc(_gclibo.GArrayUploadFile(self._gcon, c_path, c_names))
        return
            
            
    def GArrayDownloadFile(self, file_path):
        """
        Downloads a csv file containing array data at file_path.
        See Link GArrayDownloadFile() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a14b448ab8c7e6cf495865af301be398e.html#a14b448ab8c7e6cf495865af301be398e>
        """
        self._cc()
        c_path = _GCStringIn(file_path.encode(_enc))
        _rc(_gclibo.GArrayDownloadFile(self._gcon, c_path))
        return        
    
    
    def GArrayUpload(self, name, first, last):
        """
        Uploads array data from the controller's array table. 
        See Link GArrayUpload() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_af215806ec26ba06ed3f174ebeeafa7a7.html#af215806ec26ba06ed3f174ebeeafa7a7>
        """
        self._cc()
        c_name = _GCStringIn(name.encode(_enc))
        _rc(_gclib.GArrayUpload(self._gcon, c_name, first, last, 1, self._buf, _buf_size)) #1 is comma delimiter
        string_list = str(self._buf.value.decode(_enc)).split(',')
        float_list = []
        for s in string_list:
            float_list.append(float(s))
        return float_list
    
    
    def GTimeout(self, timeout):
        """
        Set the library timeout. Set to -1 to use the intitial library timeout, as specified in GOpen.
        See Link GTimeout() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a179aa2d1b8e2227944cc06a7ceaf5640.html#a179aa2d1b8e2227944cc06a7ceaf5640>
        """
        self._cc()
        _rc(_gclibo.GTimeout(self._gcon, timeout))
        self._timeout = timeout
        return
        
    
    @property
    def timeout(self):
        """
        Convenience property read access to timeout value. If -1, gclib uses the initial library timeout, as specified in GOpen.
        See Link GTimeout() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a179aa2d1b8e2227944cc06a7ceaf5640.html#a179aa2d1b8e2227944cc06a7ceaf5640>
        """
        return self._timeout
        
    @timeout.setter
    def timeout(self, timeout):
        """
        Convenience property write access to timeout value. Set to -1 to use the initial library timeout, as specified in GOpen.
        See Link GTimeout() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a179aa2d1b8e2227944cc06a7ceaf5640.html#a179aa2d1b8e2227944cc06a7ceaf5640>
        """
        self.GTimeout(timeout)
        return

        
    def GFirmwareDownload(self, file_path):
        """
        Upgrade firmware. 
        See Link GFirmwareDownload() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_a1878a2285ff17897fa4fb20182ba6fdf.html#a1878a2285ff17897fa4fb20182ba6fdf>
        """
        self._cc()
        c_path = _GCStringIn(file_path.encode(_enc))
        _rc(_gclib.GFirmwareDownload(self._gcon, c_path))
        return


    def GMessage(self):
        """
        Provides access to unsolicited messages from the controller. 
        See Link GMessage() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclib_8h_aabc5eaa09ddeca55ab8ee048b916cbcd.html#aabc5eaa09ddeca55ab8ee048b916cbcd>
        """
        self._cc()
        _rc(_gclib.GMessage(self._gcon, self._buf, _buf_size))
        return str(self._buf.value.decode(_enc))
     
     
    def GMotionComplete(self, axes):
        """
        Blocking call that returns once all axes specified have completed their motion. 
        See Link GMotionComplete() <http://www.galil.com/sw/pub/all/doc/gclib/html/gclibo_8h_a19c220879442987970706444197f397a.html#a19c220879442987970706444197f397a>
        """
        self._cc()
        c_axes = _GCStringIn(axes.encode(_enc))
        _rc(_gclibo.GMotionComplete(self._gcon, c_axes))
        return

    def GInterrupt(self):
        """   
        Provides access to PCI and UDP interrupts from the controller.
        See Link GInterrupt() <http://galil.com/sw/pub/all/doc/gclib/html/gclib_8h_a5bcf802404a96343e7593d247b67f132.html#a5bcf802404a96343e7593d247b67f132>
        """
        self._cc()
        status = _GStatus(0)
        _rc(_gclib.GInterrupt(self._gcon, byref(status)))
        return status.value
    
    def GSetupDownloadFile(self, file_path, options):
        """
    Downloads specified sectors from a Galil compressed backup (gcb) file to a controller.
    
    Returns a dictionary with the controller information stored in the gcb file.
    If options is specified as 0, an additional "options" key will be in the dictionary indicating the info sectors available in the gcb
    """
        self._cc()
        c_path = _GCStringIn(file_path.encode(_enc))

        rc = _gclibo.GSetupDownloadFile(self._gcon, c_path, options, self._buf, _buf_size)
        if (options != 0):
            _rc(rc)

        info_dict = {}
        for line in str(self._buf.value.decode(_enc)).split("\"\n"):
            fields = line.split(',',1)

            if (fields[0] == ""): continue
            elif len(fields) >= 2:
                info_dict[fields[0].strip("\"\'")] = fields[1].strip("\"\'")
            else:
                info_dict[fields[0].strip("\"\'")] = ''

        if (options == 0):
            info_dict["options"] = rc

        return info_dict


laser = "488 nm"
image_counter = 0
delay_s = 1 # give plenty of delay between snaps to avoid hickups
self.set_laser(laser, wait_until_done=True, update_etl=False)
etl_midvalues_R_L = {"488 nm": (2.5, 2.2),
					"561 nm": (2.7, 2.4)}
# R-arm offset probing
self.set_shutterconfig("Right")
self.open_shutters()
self.sig_state_request.emit({'etl_r_amplitude' : 0})
min_offset = etl_midvalues_R_L[laser][0] - 0.5
max_offset = etl_midvalues_R_L[laser][0] + 0.5
for offset in np.linspace(min_offset, max_offset, 21):
	self.sig_state_request.emit({'etl_r_offset' : offset})
	time.sleep(delay_s)
	self.snap()
	image_counter += 1
	print(image_counter)

# L-arm offset
self.set_shutterconfig("Left")
self.open_shutters()
self.sig_state_request.emit({'etl_l_amplitude' : 0})
min_offset = etl_midvalues_R_L[laser][1] - 0.5
max_offset = etl_midvalues_R_L[laser][1] + 0.5
for offset in np.linspace(min_offset, max_offset, 21):
	self.sig_state_request.emit({'etl_l_offset' : offset})
	time.sleep(delay_s)
	self.snap()
	image_counter += 1
	print(image_counter)



from PyQt5 import QtWidgets, QtGui, QtCore
import os
from .widgets import MarkPositionWidget, SliderWithValue

class ComboDelegate(QtWidgets.QItemDelegate):
    '''
    A delegate that places a fully functioning QComboBox in every
    cell of the column to which it's applied

    TODO: Provide list for entries during __init__
    '''
    def __init__(self, parent, option_list=[]):
        super().__init__(parent)
        self.option_list = option_list

    def createEditor(self, parent, option, index):
        combo = QtWidgets.QComboBox(parent)
        combo.addItems(self.option_list)
        combo.currentIndexChanged.connect(self.currentIndexChanged)
        # self.connect(combo, QtCore.SIGNAL("currentIndexChanged(int)"), self, QtCore.SLOT("currentIndexChanged()"))
        return combo

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        ''' The editor needs to know the config of the filters'''

        current_item = index.model().data(index, role=QtCore.Qt.EditRole)
        current_index = self.option_list.index(current_item)

        editor.setCurrentIndex(current_index)
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        # model.setData(index, editor.currentIndex())
        model.setData(index, editor.currentText())

    def currentIndexChanged(self):
        self.commitData.emit(self.sender())


class SliderDelegate(QtWidgets.QStyledItemDelegate):
    ''' A slider delegate '''
    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        slider = QtWidgets.QSlider(parent)
        slider.setOrientation(QtCore.Qt.Horizontal)
        slider.setTracking(False)
        slider.setAutoFillBackground(True)
        slider.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        return slider

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(int(index.model().data(index, role=QtCore.Qt.EditRole)))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.value())


class SliderWithValueDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        slider = SliderWithValue(parent)
        slider.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        slider.setValue(int(index.model().data(index, role=QtCore.Qt.EditRole)))
        slider.setText(index.model().data(index, role=QtCore.Qt.EditRole))
        return slider

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(int(index.model().data(index, role=QtCore.Qt.EditRole)))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.value())


class IntensitySpinBoxDelegate(QtWidgets.QStyledItemDelegate):
    ''' Delegate with Spinbox for laser intensity'''
    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        spinbox = QtWidgets.QSpinBox(parent)
        spinbox.setSingleStep(5)
        spinbox.setMinimum(0)
        spinbox.setMaximum(100)
        spinbox.setSuffix(' %')
        spinbox.setAlignment(QtCore.Qt.AlignRight)
        spinbox.setValue(int(index.model().data(index, role=QtCore.Qt.EditRole)))
        spinbox.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        spinbox.setAutoFillBackground(True)
        return spinbox

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(index.model().data(index, role=QtCore.Qt.EditRole))
        editor.blockSignals(False)


class ProgressBarDelegate(QtWidgets.QStyledItemDelegate):
    ''' A progress bar as a delegate

    Of course, it does not have an editing function, instead it only serves
    to display data

    TODO: How to set value properly?
    '''

    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        progressbar = QtWidgets.QProgressBar(parent)
        progressbar.setOrientation(QtCore.Qt.Horizontal)
        progressbar.setAutoFillBackground(True)
        # progressbar.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        return progressbar

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(int(index.model().data(index, role=QtCore.Qt.EditRole)))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.value())

class ETLSpinBoxDelegate(QtWidgets.QStyledItemDelegate):
    ''' Delegate with Spinbox, very fine steps'''
    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        spinbox = QtWidgets.QDoubleSpinBox(parent)
        spinbox.setDecimals(3)
        spinbox.setSingleStep(0.001)
        spinbox.setMinimum(-5.0)
        spinbox.setMaximum(5.0)
        spinbox.setSuffix(' V')
        spinbox.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        spinbox.setAutoFillBackground(True)
        return spinbox

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(index.model().data(index, role=QtCore.Qt.EditRole))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.value())

class RotationSpinBoxDelegate(QtWidgets.QStyledItemDelegate):
    ''' Delegate with Spinbox, very fine steps'''
    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        spinbox = QtWidgets.QDoubleSpinBox(parent)
        spinbox.setDecimals(1)
        spinbox.setSingleStep(1)
        spinbox.setMinimum(-360.0)
        spinbox.setMaximum(360.0)
        spinbox.setSuffix('°')
        spinbox.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        spinbox.setAutoFillBackground(True)
        return spinbox

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(index.model().data(index, role=QtCore.Qt.EditRole))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.value())        

class ZstepSpinBoxDelegate(QtWidgets.QStyledItemDelegate):
    ''' Delegate with Spinbox, Minimum value is 0 (no negative step sizes)'''
    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        spinbox = QtWidgets.QDoubleSpinBox(parent)
        spinbox.setMinimum(0.1)
        spinbox.setMaximum(1000)
        spinbox.setDecimals(1)
        spinbox.valueChanged.connect(lambda: self.commitData.emit(self.sender()))
        spinbox.setAutoFillBackground(True)
        return spinbox

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setValue(int(index.model().data(index, role=QtCore.Qt.EditRole)))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.value())


class MarkPositionDelegate(QtWidgets.QStyledItemDelegate):
    ''' Delegate with a Mark button

    Upon hitting the mark button, it should take the current value
    from e.g. the coordinates and put them into the model
    set it into

    There is only a single delegate per row instantiated.
    '''
    def __init__(self, parent):
        super().__init__(parent)

        self.model = parent.model

    def createEditor(self, parent, option, index):
        marker = MarkPositionWidget(parent)
        marker.pressed.connect(lambda: self.commitData.emit(self.sender()))
        marker.lineEdit.editingFinished.connect(lambda: self.set_model_data_from_focus_change(marker, index))
        return marker

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.lineEdit.setText(str(index.model().data(index, role=QtCore.Qt.EditRole)))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        ''' Sets the data in the model in two ways:
        * When the Button is pressed, it should use the model position
        * If the lineEdit has a value change, it should use this

        Important: This functionality has to be inheritable
        '''

        ''' By testing, I found out that editing the lineEdit
        leads to self.sender() == None (print(self.sender()))
        --> self.sender() is of NoneType

        This means that I can differentiate whether an edit action
        was caused by the button press or editing the lineEdit
        '''
       
        if self.sender() is None:
            # print(self.sender())
            self.set_model_data_from_lineedit(editor, model, index)
        else:
            self.set_model_data_from_button(editor, model, index)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)

    def set_model_data_from_button(self, editor, model, index):
        pos = 0 # replace by model.position['x_pos'] after inheriting
        pos = round(pos, 2)
        model.setData(index, pos)
        editor.lineEdit.setText(str(pos))

    def set_model_data_from_lineedit(self, editor, model, index):
        try:
            model.setData(index, float(editor.lineEdit.text()))
        except:
            model.setData(index, 0.0)

    def set_model_data_from_focus_change(self, editor, index):
        try:
            self.model.setData(index, float(editor.lineEdit.text()))
        except:
            self.model.setData(index, 0.0)

class MarkXPositionDelegate(MarkPositionDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def set_model_data_from_button(self, editor, model, index):
        pos = model.state['position']['x_pos']
        pos = round(pos, 2)
        model.setData(index, pos)
        editor.lineEdit.setText(str(pos))

class MarkYPositionDelegate(MarkPositionDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def set_model_data_from_button(self, editor, model, index):
        pos = model.state['position']['y_pos']
        pos = round(pos, 2)
        model.setData(index, pos)
        editor.lineEdit.setText(str(pos))

class MarkZPositionDelegate(MarkPositionDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def set_model_data_from_button(self, editor, model, index):
        pos = model.state['position']['z_pos']
        pos = round(pos, 2)
        model.setData(index, pos)
        editor.lineEdit.setText(str(pos))

class MarkFocusPositionDelegate(MarkPositionDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def set_model_data_from_button(self, editor, model, index):
        pos = model.state['position']['f_pos']
        pos = round(pos, 2)
        model.setData(index, pos)
        editor.lineEdit.setText(str(pos))

class MarkRotationPositionDelegate(MarkPositionDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def set_model_data_from_button(self, editor, model, index):
        pos = model.state['position']['theta_pos']
        pos = round(pos, 2)
        model.setData(index, pos)
        editor.lineEdit.setText(str(pos))

class ChooseFolderDelegate(QtWidgets.QItemDelegate):
    '''
    A delegate that has a button, in turn opening a set folder dialog 
    and displaying the folder name.

    '''
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.folder = ''

    def createEditor(self, parent, option, index):
        button = QtWidgets.QPushButton(parent)
        button.setText('...')
        button.clicked.connect(self.choose_folder_dialog)
        return button
    
    def choose_folder_dialog(self):
        if os.path.isdir(self.parent.cfg.startup['folder']):
            ini_folder = self.parent.cfg.startup['folder']
        else:
            ini_folder = ''
            print(f"Warning: 'folder' specified in config file does no exist: {self.parent.cfg.startup['folder']}")

        path = QtWidgets.QFileDialog.getExistingDirectory(self.parent, 'Select Folder', ini_folder)
        if path:
            self.folder = path
            self.commitData.emit(self.sender())

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        editor.setText(str(index.model().data(index, role=QtCore.Qt.EditRole)))
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        model.setData(index, self.folder)



"""
mesoSPIM Module for creating waveforms and analog output signals

Author: Fabian Voigt

#TODO
* Usage of amplitude is not consistent (peak-to-peak in single_pulse)
"""

#import nidaqmx
# from nidaqmx.constants import AcquisitionType, TaskMode
# from nidaqmx.constants import LineGrouping

from scipy import signal
import numpy as np

def single_pulse(
    samplerate=100000,  # in samples/second
    sweeptime=0.4,      # in seconds
    delay=10,           # in percent
    pulsewidth=1,       # in percent
    amplitude=0,        # in volts
    offset=0            # in volts
    ):

    '''
    Returns a numpy array with a single pulse

    Used for creating TTL pulses out of analog outputs and laser intensity
    pulses.

    Units:
    samplerate: Integer
    sweeptime:  Seconds
    delay:      Percent
    pulsewidth: Percent
    amplitude:  Volts
    offset:     Volts

    Examples:

    typical_TTL_pulse = single_pulse(samplerate, sweeptime, 10, 1, 5, 0)
    typical_laser_pulse = single_pulse(samplerate, sweeptime, 10, 80, 1.25, 0)
    '''

    # get an integer number of samples
    samples = int(np.floor(np.multiply(samplerate, sweeptime)))
    # create an array just containing the offset voltage:
    array = np.zeros((samples))+offset

    # convert pulsewidth and delay in % into number of samples
    pulsedelaysamples = int(samples * delay / 100)
    pulsesamples = int(samples * pulsewidth / 100)

    # modify the array
    array[pulsedelaysamples:pulsesamples+pulsedelaysamples] = amplitude
    return np.array(array)

def tunable_lens_ramp(
    samplerate = 100000,    # in samples/second
    sweeptime = 0.4,        # in seconds
    delay = 7.5,            # in percent
    rise = 85,              # in percent
    fall = 2.5,             # in percent
    amplitude = 0,          # in volts
    offset = 0              # in volts
    ):

    '''
    Returns a numpy array with a ETL ramp

    The waveform starts at offset and stays there for the delay period, then
    rises linearly to 2x amplitude (amplitude here refers to 1/2 peak-to-peak)
    and drops back down to the offset voltage during the fall period.

    Switching from a left to right ETL ramp is possible by exchanging the
    rise and fall periods.

    Units of parameters
    samplerate: Integer
    sweeptime:  Seconds
    delay:      Percent
    rise:       Percent
    fall:       Percent
    amplitude:  Volts
    offset:     Volts
    '''
    # get an integer number of samples
    samples = int(np.floor(np.multiply(samplerate, sweeptime)))
    # create an array just containing the negative amplitude voltage:
    array = np.zeros((samples))-amplitude + offset

    # convert rise, fall, and delay in % into number of samples
    delaysamples = int(samples * delay / 100)
    risesamples = int(samples * rise / 100)
    fallsamples = int(samples * fall / 100)

    risearray = np.arange(0,risesamples)
    risearray = amplitude * (2 * np.divide(risearray, risesamples) - 1) + offset

    fallarray = np.arange(0,fallsamples)
    fallarray = amplitude * (1-2*np.divide(fallarray, fallsamples)) + offset

    # rise phase
    array[delaysamples:delaysamples+risesamples] = risearray
    # fall phase
    array[delaysamples+risesamples:delaysamples+risesamples+fallsamples] = fallarray

    return np.array(array)

def sawtooth(
    samplerate = 100000,    # in samples/second
    sweeptime = 0.4,        # in seconds
    frequency = 10,         # in Hz
    amplitude = 0,          # in V
    offset = 0,             # in V
    dutycycle = 50,          # dutycycle in percent
    phase = np.pi/2,          # in rad
    ):
    '''
    Returns a numpy array with a sawtooth function

    Used for creating the galvo signal.

    Example:
    galvosignal =  sawtooth(100000, 0.4, 199, 3.67, 0, 50, np.pi)
    '''

    samples =  int(samplerate*sweeptime)
    dutycycle = dutycycle/100       # the signal.sawtooth width parameter has to be between 0 and 1
    t = np.linspace(0, sweeptime, samples)
    # Using the signal toolbox from scipy for the sawtooth:
    waveform = signal.sawtooth(2 * np.pi * frequency * t + phase, width=dutycycle)
    # Scale the waveform to a certain amplitude and apply an offset:
    waveform = amplitude * waveform + offset

    return waveform

def square(
    samplerate = 100000,    # in samples/second
    sweeptime = 0.4,        # in seconds
    frequency = 10,         # in Hz
    amplitude = 0,          # in V
    offset = 0,             # in V
    dutycycle = 50,         # dutycycle in percent
    phase = np.pi,          # in rad
    ):
    """
    Returns a numpy array with a rectangular waveform
    """

    samples =  int(samplerate*sweeptime)
    dutycycle = dutycycle/100       # the signal.square duty parameter has to be between 0 and 1
    t = np.linspace(0, sweeptime, samples)

    # Using the signal toolbox from scipy for the sawtooth:
    waveform = signal.square(2 * np.pi * frequency * t + phase, duty=dutycycle)
    # Scale the waveform to a certain amplitude and apply an offset:
    waveform = amplitude * waveform + offset

    return waveform
